#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>

#include <stdio.h>
#include <stdlib.h>

#include "usart.h"

#define TIMER1_TOP 0xFF
#define SPEED1 0xFF
#define SPEED2 0xD0 // SPEED1 * 85%
#define SPEED3 0xBF // SPEED1 * 75%
#define SPEED4 0x90 // SPEED1 * 60%
#define SPEED5 0x7F // SPEED1 * 50%

#define END_CONN_CODE '0'
#define FORWARD_CODE '1'
#define BACKWARD_CODE '2'
#define LEFT_CODE '3'
#define RIGHT_CODE '4'
#define STOP_CODE '5'
#define TOGGLE_LIGHTS_CODE '6'
#define TOGGLE_L_LIGHTS_CODE '7'
#define TOGGLE_R_LIGHTS_CODE '8'
#define SPEED_UP_CODE '9'

#define TRUE 1
#define FALSE 0

/*
 * Motor Driver1:
 * IN1 -> PC3
 * IN2 -> PC5
 * IN3 -> PC6
 * IN4 -> PC4
 *
 * ENA -> PB3
 * ENB -> PD7
 */
/*
 * Motor Driver2:
 * IN1 -> PA3
 * IN2 -> PA2
 * IN3 -> PA1
 * IN4 -> PA0
 *
 * ENA -> PB4
 * ENB -> PD6
 *
 * Right Lights -> PC1
 * Left Lights -> PC2
 */

static volatile unsigned char command, prev_cmd;
static volatile unsigned char curr_speed;
static volatile unsigned char right_lights, left_lights, all_lights;
static volatile unsigned char new_cmd = 0;
static volatile unsigned char last_move;

ISR(USART0_RX_vect) {
	new_cmd = TRUE;
	prev_cmd = command;
	command = UDR0;
}

ISR(TIMER1_OVF_vect) {
	if (right_lights) {
		PORTC ^= (1 << PC1);
	}
	if (left_lights) {
		PORTC ^= (1 << PC2);
	}
}

void forward(unsigned char speed) {
	OCR0A = speed;
	OCR2A = speed;
	OCR0B = speed;
	OCR2B = speed;

	PORTC |= (1 << PC3);
	PORTC &= ~(1 << PC5);
	PORTA |= (1 << PA2);
	PORTA &= ~(1 << PA3);

	PORTC |= (1 << PC4);
	PORTC &= ~(1 << PC6);
	PORTA |= (1 << PA1);
	PORTA &= ~(1 << PA0);
}

void backward(unsigned char speed) {
	OCR0A = speed;
	OCR2A = speed;
	OCR0B = speed;
	OCR2B = speed;

	PORTC |= (1 << PC5);
	PORTC &= ~(1 << PC3);
	PORTA |= (1 << PA3);
	PORTA &= ~(1 << PA2);

	PORTC |= (1 << PC6);
	PORTC &= ~(1 << PC4);
	PORTA |= (1 << PA0);
	PORTA &= ~(1 << PA1);
}

void left(unsigned char speed) {
	/*
	 * We lower the value for the left engine's speed
	 */
	switch (speed) {
		case SPEED1 : OCR2A = SPEED3; OCR2B = SPEED3; break;
		case SPEED2 : OCR2A = SPEED4; OCR2B = SPEED4; break;
		case SPEED3 : OCR2A = SPEED4; OCR2B = SPEED4; break;
		default: break;
	}
}

void right(unsigned char speed) {
	/*
	 * We lower the value for the right engine's speed
	 */
	switch (speed) {
		case SPEED1 : OCR0A = SPEED3; OCR0B = SPEED3; break;
		case SPEED2 : OCR0A = SPEED4; OCR0B = SPEED4; break;
		case SPEED3 : OCR0A = SPEED4; OCR0B = SPEED4; break;
		default: break;
	}
}

void stop() {
	PORTC &= ~((1 << PC3) | (1 << PC4) | (1 << PC5) | (1 << PC6));
	PORTA &= ~((1 << PA0) | (1 << PA1) | (1 << PA2) | (1 << PA3));
	OCR0A = 0;
	OCR2A = 0;
	OCR0B = 0;
	OCR2B = 0;
}

void timer0_init() {
	/*
	 * Timer0 - Channel A (OC0A) and Channel B (OC0B)
	 */
	DDRB |= (1 << PB3) | (1 << PB4);

	TCCR0A &= 0;
	TCCR0B &= 0;
	/*
	 * FastPWM - 0xFF; clear on comp, set on bottom
	 * PS = 8; f ~ 8kHz
	 */
	TCCR0A |= (1 << WGM00) | (1 << WGM01) | (1 << COM0A1) | (1 << COM0B1);
	TCCR0B |= (1 << CS01);

}

void timer2_init() {
	/*
	 * Timer2 - Channel A (OC2A) [PD7] and Channel B (OC2B) [PD6]
	 */
       	DDRD |= (1 << PD6) | (1 << PD7);
	
	TCCR2A &= 0;
 	TCCR2B &= 0;

	/*
	 * Like above
	 */
	TCCR2A |= (1 << WGM20) | (1 << WGM21) | (1 << COM2A1) | (1 << COM2B1);
	TCCR2B |= (1 << CS21);	
}

void timer1_init() {
	/*
	 * Timer1 - Channel A (OC1A) [PD5]
	 */
	DDRD |= (1 << PD5);

	TCCR1A &= 0;
	TCCR1B &= 0;
	/*
	 * Normal - PS = 256
	 */
	TCCR1A |= (1 << COM1A1);
	TCCR1B |= (1 << CS12);
	/*
	 *TOIE1 to be set
	*/
}

void engines_init() {
	/*
	 * Setting the above-mentioned pins as output
	 */
	
	DDRC &= 0;
	DDRA &= 0;
	DDRC |= (1 << PC3) | (1 << PC4) | (1 << PC5) | (1 << PC6);
	DDRA |= (1 << PA0) | (1 << PA1) | (1 << PA2) | (1 << PA3);
	/*
	 * Initially, all engines are stopped
	 */
	PORTC &= ~((1 << PC3) | (1 << PC4) | (1 << PC5) | (1 << PC6));
	PORTA &= ~((1 << PA0) | (1 << PA1) | (1 << PA2) | (1 << PA3));
}

void leds_init() {
	DDRC &= 0;
	DDRC |= (1 << PC2) | (1 << PC1);
	/*
	 * Initially, all lights are off
	 */
	PORTC &= ~((1 << PC2) | (1 << PC1));
}

void main_logic() {
	while (true) {
		if (new_cmd) {
			switch (command) {
			case END_CONN_CODE : return;
			case FORWARD_CODE : forward(curr_speed);
				last_move = FORWARD_CODE;
	       			new_cmd = FALSE;
				break;
			case BACKWARD_CODE : backward(curr_speed);
				last_move = BACKWARD_CODE;
	       			new_cmd = FALSE;
				break;
			case LEFT_CODE : left(curr_speed);
				last_move = LEFT_CODE;
	       			new_cmd = FALSE;
				break;
			case RIGHT_CODE : right(curr_speed);
				last_move = RIGHT_CODE;
	       			new_cmd = FALSE;
				break;
			case STOP_CODE : stop();
				new_cmd = FALSE;
				break;
			case TOGGLE_LIGHTS_CODE : 
				 	if (all_lights) {
				 		PORTC &= ~((1 << PC1) | (1 << PC2));
						all_lights = 0;
					} else {
						if (right_lights || left_lights) {
							right_lights = 0;
							left_lights = 0;
							TIMSK1 &= ~(1 << TOIE1);
						}
						PORTC |= (1 << PC1) | (1 <<PC2);
						all_lights = 1;
					}
					new_cmd = FALSE;
					break;
			case TOGGLE_R_LIGHTS_CODE :
					if (right_lights) {
					  	right_lights = 0;
						if (!left_lights)
					  		TIMSK1 &= ~(1 << TOIE1);
						PORTC &= ~(1 << PC1);
					} else {
						if (all_lights)
							PORTC &= ~(1 << PC1);
						right_lights = 1;
						TIMSK1 |= (1 << TOIE1);
					}
					new_cmd = FALSE;
					break;

			case TOGGLE_L_LIGHTS_CODE :
					if (left_lights) {
						left_lights = 0;
						if (!right_lights) {
							TIMSK1 &= ~(1 << TOIE1);
						}
						PORTC &= ~(1 << PC2);
					} else {
						if (all_lights)
							PORTC &= ~(1 << PC2);
						left_lights = 1;
						TIMSK1 |= (1 << TOIE1);
					}
					new_cmd = FALSE;
					break;
			case SPEED_UP_CODE :
				if (curr_speed == SPEED3)
			       		curr_speed = SPEED2;
				else
					curr_speed = SPEED1;
				command = last_move;
				break;	
			default : continue;
			}
		}
	}
}

int main() {
	sei();
	engines_init();
	timer0_init();
	timer1_init();
	timer2_init();

	USART0_init();
	leds_init();
	all_lights = 0;
	left_lights = 0;
	right_lights = 0;

	curr_speed = SPEED1;

	main_logic();

	return 0;
}
